#!/usr/bin/env python
from __future__ import print_function
from argtools import command, argument
import logging
import numpy as np
import itertools
from collections import namedtuple, defaultdict
from itertools import groupby
from builtins import zip
from multiprocess import Pool
import signal
from scipy.stats import norm
from ..covfit import load_gender_info, get_ploidies
#from ..covfit import gmm
from ..covfit import gmm_v2 as gmm
from ..utils import iter_tabs, Counter, with_header


@command.add_sub
@argument('coverage_bed')
@argument('--col-region-id')
@argument('--col-sample-id', default='sampleid')
@argument('--col-cov', default='mean_cov')
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('-m', '--max-copy', type=int, default=10)
@argument('-la', type=float, default=0.3, help='penalty of offset')
@argument('-lb', type=float, default=1.0, help='penalty of ratio')
@argument('-lc', type=float, default=0.3, help='penalty of precision')
@argument('-a', nargs='*', type=float, default=[0., 0.4, 0.8], help='penalty of offset')
@argument('-b', nargs='*', type=float, default=[1.0], help='penalty of ratio')
@argument('-c', type=float, default=1.0, help='penalty of precision')
@argument('-f', '--max-diff', type=float, default=1e-6, help='max diff of log likelihood')
@argument('-p', '--process', type=int, default=8)
@argument('--delta', type=float, default=0.3, help='variance for zero copy')
@argument('--step-limit', type=int, help='the number of steps until inference stopped')
@argument('--gender-info', help='col1: sample_id, col2: {1:male, 2:female}')
def covfit_gmm(args):
    """

    1. chrom
    2. start
    3. end
    4. region id
    """
    if args.gender_info:
        sample_genders = load_gender_info(args.gender_info)
    else:
        sample_genders = defaultdict(lambda : 1)
    zoom = 2. / args.diploid_value

    # not required
    def imap_wrap(pool, task, task_inputs, chunksize=1):  # test
        result = pool.imap(task, task_inputs, chunksize=1)
        while 1:
            yield result.next(timeout=999999)


    def run():
        def init_pool():
            logging.info('Init pool')
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        pool = Pool(args.process, init_pool)
        try:
            task_inputs = gen_inputs()

            if args.process == 1:
                results = itertools.imap(task, task_inputs)
            else:
                results = pool.imap(task, task_inputs, 1)
                #results = imap_wrap(pool, task, task_inputs, 100)

            data_header = ['chrom', 'start', 'end', 'region_id', 'nsamples']
            param_header = ['status', 'ia', 'ib', 'ic', 'a', 'b', 'c', 'top_cn', 'top_theta', 'theta', 'll', 'step']
            out_header = data_header + param_header
            print (*out_header, sep='\t')
            for data, params in results:
                row = [data[c] for c in data_header] \
                    + [params[c] for c in param_header]
                print (*row, sep='\t')
            pool.close()
        except Exception as e:
            pool.terminate()
            raise e
        finally:
            pool.join()

    def gen_inputs():
        with open(args.coverage_bed) as fp:
            it = iter_tabs(fp)
            header = next(it)
            idx_region_id = header.index(args.col_region_id) if args.col_region_id else 3
            idx_sample_id = header.index(args.col_sample_id)
            idx_cov = header.index(args.col_cov)

            for region_id, rows in groupby(it, lambda x: x[idx_region_id]):
                rows = list(rows)
                chrom = rows[0][0]

                samples = []
                genders = []
                covs = []

                for row in rows:
                    sample_id = row[idx_sample_id]
                    cov = float(row[idx_cov])
                    if np.isnan(cov):
                        continue
                    if cov * zoom > args.max_copy:  # coverage larger than the max copy is not used for parameter estimation (adhoc treatment)
                        continue
                    samples.append(sample_id)
                    genders.append(sample_genders[sample_id])
                    covs.append(cov)

                ploidies = get_ploidies(chrom, genders)
                ploidy_counts = Counter(ploidies)

                data = dict(
                        chrom = rows[0][0],
                        start = int(rows[0][1]),
                        end = int(rows[0][2]),
                        region_id = region_id,
                        covs = covs,
                        zoom = zoom,
                        la = args.la,
                        lb = args.lb,
                        lc = args.lc,
                        a = args.a,
                        b = args.b,
                        c = args.c,
                        delta = args.delta,
                        max_diff = args.max_diff,
                        step_limit = args.step_limit,
                        ploidies = ploidies,
                        max_copy = args.max_copy,
                        nsamples = ','.join([str(ploidy_counts[cn]) for cn in (0, 1, 2)]),
                )

                yield data

    def task(data):
        #signal.signal(signal.SIGINT, signal.SIG_IGN)
        chrom = data['chrom']
        start = data['start']
        end = data['end']
        name = data['region_id']
        ploidies = data['ploidies']
        init_as = data['a']
        init_bs = data['b']
        init_abs = [(a, b) for a in init_as for b in init_bs]
        init_c = data['c']
        la = data['la']
        lb = data['lb']
        lc = data['lc']
        covs = np.array(data['covs'])
        zoom = data['zoom']
        max_diff = data['max_diff']
        step_limit = data['step_limit']
        max_copy = data['max_copy']

        mod_depths = (covs * zoom)
        #ploidy_counts = Counter(ploidies)  # 0, 1, 2
        #nsamples_str =  ','.join(str(ploidy_counts[i]) for i in (0, 1, 2))

        n_callable = len(covs)
        logging.info('[%s] n_callable: %s', name, n_callable)
        #init_as = (0.,)
        results = []
        for init_a, init_b in init_abs:
            norm = None
            try:
                assert n_callable > 0, 'The number of callable samples ({0}) is less than the minimum value'.format(n_callable)
                norm = gmm.Normalizer(mod_depths, ploidies=ploidies, max_copy=max_copy, a=init_a, b=init_b, c=init_c, la=la, lb=lb, lc=lc, max_diff=max_diff, step_limit=step_limit, name=name)
                #init_a, init_b, _, _ = get_best_fit(mod_depths)
                result = norm.iteration()
                top_cn = np.argmax(result.theta)
                params = {
                        'status': 1,
                        'ia': init_a,
                        'ib': init_b,
                        'ic': init_c,
                        'a': result.a,
                        'b': result.b,
                        'c': result.c,
                        'theta': ','.join(map(str, result.theta)),
                        'step': result.step,
                        'top_cn': top_cn,
                        'top_theta': result.theta[top_cn],
                        'll': result.ll,
                }
            except AssertionError as e:
                logging.warning('Assert error: %s', e)
                logging.warning('Cannot optimize model for %s', data)
                step = 0 if norm is None else norm.step
                params = {
                        'status': 0,
                        'ia': init_a,
                        'ib': init_b,
                        'ic': init_c,
                        'a': init_a,
                        'b': init_b,
                        'c': init_c,
                        'theta': '',
                        'step': step,
                        'top_cn': 0,
                        'top_theta': 0,
                        'll': 0,
                }
            results.append(params)
        best_params = max(results, key=lambda v: (v['status'], v['ll']))

        logging.info('Emitting %s:%s-%s, %s, %s', chrom, start, end, name, best_params)
        return (data, best_params)

    run()


@command.add_sub
@argument('fit_result')
@argument('--min-call-nsample', type=int, default=1)
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('--delta', type=float, default=0.3, help='variance for zero copy')    # TODO incorporate into the result
@argument('-t', '--threshold', type=float, default=.8)
@argument('--sex-info', help='list file of 1:male, 2:female')
def covfit_gmm_genotype(args):
    """
    """
    nsample = args.nsample
    zoom = 2. / args.diploid_value
    gender_info = [int(line.rstrip()) for line in open(gender_info)]
    get_ploidies = get_ploidy_factory(nsample, gender_info)

    for row in iter_tabs(open(args.fit_result)):
        chrom = row[0]
        start = int(row[1])
        end = int(row[2])
        name = row[3]

        depths = np.array(row[-nsample:], dtype=float)
        callable_idxs = np.where(~ np.isnan(depths))[0]
        n_callable = len(callable_idxs)

        mod_depths = (depths * zoom)[callable_idxs]

        #mod_depths = depths * zoom
        keys = row[-args.nsample-2].split(':')
        params = dict(zip(keys, row[-args.nsample-1].split(':')))
        a = float(params['a'])
        b = float(params['b'])
        c = float(params['c'])
        try:
            assert n_callable >= args.min_call_nsample, 'The number of callable samples ({0}) is less than the minimum value'.format(n_callable)
            theta = np.array(map(float, params['theta'].split(',')))
            max_copy = len(theta) - 1

            ploidies = tuple(np.array(get_ploidies(chrom))[callable_idxs])

            #ploidies = get_ploidies(chrom)
            #ploidies = np.array(ploidies)

            norm = gmm.Normalizer(mod_depths, ploidies=ploidies, max_copy=max_copy, a=a, b=b, c=c, theta=theta, name=name, delta=delta)
            gts = list(norm.iter_cns())
            gts1 = []   # genotyped copy numbers
            gts2 = []   # genotype probability of 0/1/2
            for prob in gts:
                max_cn = np.argmax(prob)
                if prob[max_cn] >= args.threshold:
                    gts1.append(max_cn)
                else:
                    gts1.append('.')
                gts2.append('{0:.1f}:{1:.1f}:{2:.1f}'.format(prob[0], prob[1], prob[2]))
        except (ValueError, AssertionError):
            gts1 = ['.'] * nsample
            gts2 = ['.:.:.'] * nsample

        out_row = list(row[:-args.nsample-2])
        out_row.append(a)
        out_row.append(b)
        out_row.append(c)
        out_row.append(params['theta'])
        out_row.append(','.join(map(str, gts1)))
        out_row.append(','.join(map(str, gts2)))
        print (*out_row, sep='\t')


@command.add_sub
@argument('coverage')
@argument('-p', '--fit-params', required=True)
@argument('--min-valid-ratio', type=float, default=.5)
@argument('--min-valid-len', type=int, default=100)
#@argument('--col-region-id')
@argument('--col-cov', default='mean_cov', help='column name of normalized coverage')
@argument('--delta', type=float, default=0.3, help='variance for zero copy')    # TODO incorporate into the result
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('-t', '--threshold', type=float, default=.8)
@argument('--sex', type=int, choices=[1, 2], required=True)
def covfit_gmm_genotype1(args):
    """
    gt        # filtered by specified condition
    map_gt    # 
    likelihoods
    """
    zoom = 2. / args.diploid_value
    cov_it = with_header(iter_tabs(open(args.coverage)), use_header=True)
    fit_it = with_header(iter_tabs(open(args.fit_params)), use_header=True)
    cov_header = next(cov_it)
    fit_header = next(fit_it)
    #col_region_id = args.col_region_id if args.col_region_id else 'name'
    header = list(cov_header) + ['copy_number', 'top_cn', 'top_gt', 'top_gt_prob', 'gts', 'cn_prob0', 'cn_prob1', 'cn_prob2', 'cn_probs', 'cn_probs']
    print (*header, sep='\t')

    NA = float('nan')
    for cov_row, params in zip(cov_it, fit_it):
        assert all(cov_row[col] == params[col] for col in ('chrom', 'start', 'end')), 'chrom, start, and end should be aligned for the two input files'
        chrom = params['chrom']
        start = int(params['start'])
        end = int(params['end'])
        name = params['region_id']

        depth = float(cov_row[args.col_cov])
        mod_depth = depth * zoom
        ploidy = get_ploidies(chrom, [args.sex])[0]

        try:
            a = float(params['a'])
            b = float(params['b'])
            c = float(params['c'])
            theta = np.array(map(float, params['theta'].split(',')))
        except ValueError:
            theta = NA

        cn_probs = (NA,)
        low_cn_probs = (NA,) * 3
        gt_keys = ['.']
        gt_vals = [1.]
        if ploidy == 0:
            top_cn = 0
            cn = 0
            gt_keys = ['0']
            gt_vals = [1.]
            low_cn_probs = (1., 0., 0.)
        elif not np.isnan(depth) and not np.any(np.isnan(theta)):
            max_copy = len(theta) - 1
            norm = gmm.Normalizer(np.array([mod_depth]), ploidies=[ploidy], max_copy=max_copy, a=a, b=b, c=c, theta=theta, name=name, delta=args.delta)
            cn_probs = list(norm.iter_cns())[0]
            top_cn = np.argmax(cn_probs)
            cn = top_cn if cn_probs[top_cn] >= args.threshold else -1

            low_cn_probs = cn_probs[0], cn_probs[1], cn_probs[2]

            valid_ratio = float(cov_row['valid_ratio'])
            valid_len = int(cov_row['valid_len'])
            if valid_ratio < args.min_valid_ratio \
               or valid_len < args.min_valid_len:
                cn = -1


            gts = list(norm.iter_genotypes())[0]
            gt_keys, gt_vals = get_allelotype_records(gts)
            # get at records
        else:
            top_cn = -1
            cn = -1

        out_row = list(cov_row[col] for col in cov_header)
        out_row.append(cn)
        out_row.append(top_cn)
        out_row.append(gt_keys[0])
        out_row.append('{0:.4f}'.format(gt_vals[0]))
        out_row.append(':'.join(gt_keys))
        out_row.append(':'.join(map('{0:.4f}'.format, gt_vals)))
        out_row.extend(tuple(map('{0:.4f}'.format, low_cn_probs)))
        out_row.append(':'.join(map('{0:.4f}'.format, cn_probs)))
        print (*out_row, sep='\t')


def get_allelotype_records(gts, psum_max=.99):
    """
    gts: {'0': 1.0}
         {'1/0': 0.8, '2/0': 0.1}
    """
    gt_keys = []
    gt_vals = []
    psum = .0
    for key, p in sorted(gts.items(), key=lambda x: x[1], reverse=True):
        gt_keys.append(key)
        gt_vals.append(p)
        psum += p
        if psum > psum_max:
            break
    return gt_keys, gt_vals



from __future__ import print_function
from argtools import command, argument
import pysam
from pysam import Samfile
import re
import logging
from builtins import filter, zip
from ..utils import fill_text
from ..bioseq import color_term_dna, get_aln_pos_text, dna_revcomp
from ..samutil import LocalInfo, sam_intervals, Read, clip_seq
from operator import attrgetter


@command.add_sub
@argument('bamfile')
@argument('bamout')
@argument('-u', '--unpaired')
@argument('-f', '--uncount-flag', type=lambda x: int(x, 16), default=0x900, help='uncount-flag for existing reads; default is secondary or supplementary reads')
def bam_discard_unpaired(args):
    sam = pysam.Samfile(args.bamfile)
    read1s = set()
    read2s = set()
    uncount_flag = args.uncount_flag
    for rec in sam:
        if rec.flag & uncount_flag:
            continue
        if rec.is_read1:
            read1s.add(rec.qname)
        else:
            read2s.add(rec.qname)

    paired = read1s & read2s
    unpaired = (read1s - paired) | (read2s - paired)
    logging.info('Paired reads: %s', len(paired))
    logging.info('Unpaired reads: %s', len(unpaired))

    sam.reset()
    out = pysam.Samfile(args.bamout, mode='wb', template=sam)
    out2 = pysam.Samfile(args.unpaired, mode='wb', template=sam) if args.unpaired else None
    for rec in sam:
        if rec.qname in paired:
            # workaround for avoiding duplicated emission of the same record
            if rec.is_read1:
                if rec.qname in read1s:
                    out.write(rec)
                    read1s.remove(rec.qname)
            else:
                if rec.qname in read2s:
                    out.write(rec)
                    read2s.remove(rec.qname)
        elif out2:
            out2.write(rec)


@command.add_sub
@argument('bamfile')
@argument('-l', '--list')
@argument('--max-edit', type=int, help='use NM tag')
@argument('-o', '--output', default='/dev/stdout')
def bam_filter(args):
    sam = pysam.Samfile(args.bamfile)
    if args.output.endswith('.bam'):
        mode = 'wb'
    else:
        mode = 'wh'
    out = pysam.Samfile(args.output, mode=mode, template=sam)

    if args.list:
        white_list = set([name.rstrip() for name in open(args.list)])
        def list_filter(sam):
            for rec in sam:
                if rec.qname in white_list:
                    yield rec

        sam = list_filter(sam)

    if args.max_edit is not None:
        max_edit = args.max_edit
        def is_ok(rec):
            tag = dict(rec.get_tags())
            edit = tag.get('NM', 0)
            return edit <= max_edit
        sam = filter(is_ok, sam)

    for rec in sam:  # fetch only primary and non-supplementary reads
        out.write(rec)


@command.add_sub
@argument('bamfile')
@argument('-r', '--region')
def bam_profile(args):
    """
    """
    sam = pysam.Samfile(args.bamfile)
    attrs = ['name', 'read1', 'unmapped', 'suppl', 'num_frags', 'qlen', 'rname', 'start', 'end', 'alen', 'mapq', 'reverse', 'lclip', 'rclip', 'edit', 'nins', 'ndel']
    getter = attrgetter(*attrs)
    with sam:
        print (*(attrs + ['tags_md']), sep='\t')
        for rec in sam.fetch(args.region):
            r = Read(rec)
            print (*(getter(r) + (dict((k, v) for k, v in rec.tags if k != 'MD'),)), sep='\t')


# TODO sort by read_name, read_length, ref_identity, ref_diversity <= weighted identity (using base probability as weight is ok)
# show read properties (mismatch on reads
@command.add_sub
@argument('bam')
@argument('region', nargs='+')  # TODO currently, need this option
@argument('-r', '--reference')
@argument('-c', '--color-base', action='store_true')
@argument('-m', '--mask-ref-base', action='store_true')
@argument('--order', help='sort orders: select from qname, edit, nins, ndel, alen, <ref_pos>'.format())
@argument('-R', '--reverse', action='store_true')
def bam_aln_view(args):
    """
    """
    sam = Samfile(args.bam)
    skip_flags = 0x004
    sep = ' ' * 10
    sep = '\t'
    def decorate(aln, ref_aln=None, indicator=None):
        if ref_aln and args.mask_ref_base:
            #aln = ''.join(('.' if a == r and a != ('-' or ' ') else a) for a, r in zip(aln, ref_aln))
            aln = ''.join(('.' if a == r else a) for a, r in zip(aln, ref_aln))  # '-' is also shown as '.'
        if indicator:
            # if aln is blank, fill with indicator instead
            aln = ''.join(i if a == ' ' else a for a, i in zip(aln, indicator))
        if args.color_base:
            aln = color_term_dna(aln)
        return aln

    def get_sorted(read_alns):
        if args.order is None:
            return read_alns
        # if args.order.isdigit():  # TODO getting corresponding reference position is required
        #     pos = int(args.order)
        #     read_alns = sorted(read_alns, key=lambda x: x[1], reversed=args.reverse)
        order = args.order
        assert order in 'qname edit edit_ratio nins ndel alen'.split(' ')
        return sorted(read_alns, key=lambda x: getattr(x[0], order), reverse=args.reverse)

    logging.info('Target region: %s', args.region)
    with sam:
        for iv in sam_intervals(sam, regions=(args.region or None)):
            loc = LocalInfo(sam, iv, fasta=args.reference, skip_flags=skip_flags)
            print ('start', loc.start)
            print ('end', loc.end)
            print ('align length', loc.align_length)
            ref_aln = ''.join(loc.get_ref_align())
            ref_dec = decorate(ref_aln)
            pos_txt = get_aln_pos_text(ref_aln, offset=loc.start)
            indicator = pos_txt['indicator']
            print (pos_txt['number'])
            print (indicator)
            print (ref_dec, iv.contig, 'mapq', 'clip', 'edit', 'nins', 'ndel', 'alen', 'edit_ratio', sep=sep)
            indicator = indicator.replace('|', ':')  # for visibility
            print (indicator)
            it = loc.iter_read_aligns()
            it = get_sorted(it)
            for read, aln in it:
                read_aln = decorate(''.join(aln), ref_aln, indicator=indicator)
                clip_status = ('1' if read.has_lclip else '0') + ('1' if read.has_rclip else '0')
                print (read_aln, read.rec.qname, read.mapq, clip_status, read.edit, read.nins, read.ndel, read.alen,
                        '{0:.2f}'.format(read.edit_ratio), sep=sep)


@command.add_sub
@argument('bam')
@argument('-r', '--region', help='region of target bam file')
@argument('-s', '--source-bam')  # TODO pair of fastq
@argument('-o', '--output', default='/dev/stdout')
def bam_fill_seq(args):
    """ Fill empty sequence with known seqs
    """
    if not args.source_bam:
        source_bam = args.bam
    else:
        source_bam = args.source_bam
    logging.info('Loading samfile: %s', source_bam)
    src_seqs = {1: {}, 2: {}}

    src = pysam.Samfile(source_bam)
    with src:
        for rec in src:
            if rec.is_supplementary:  # skip supplementary alignment
                continue
            if rec.is_secondary:  # skip supplementary alignment
                continue
            if rec.seq is None:  # empty
                continue
            if rec.is_read1:
                src_seqs[1][rec.qname] = (rec.seq, rec.is_reverse)
            else:
                src_seqs[2][rec.qname] = (rec.seq, rec.is_reverse)

    logging.info('Loaded read1 : %s', len(src_seqs[1]))
    logging.info('Loaded read2 : %s', len(src_seqs[2]))

    sam = Samfile(args.bam)
    if args.output.endswith('.bam'):
        mode = 'wb'
    else:
        mode = 'wh'
    out = pysam.Samfile(args.output, mode=mode, template=sam)

    if args.region:
        it = sam.fetch(region=args.region)
    else:
        it = sam

    for rec in it:
        qname = rec.qname
        if rec.seq is None:  # only fill when empty
            ret = src_seqs[1 if rec.is_read1 else 2].get(rec.qname)
            if ret is not None:
                seq, is_rev = ret
                if is_rev != rec.is_reverse:
                    seq = dna_revcomp(seq)
                seq = clip_seq(seq, rec.cigarstring)
                rec.seq = seq  # refill

        out.write(rec)



from argtools import command, argument
import logging
import os
import re
import math
from .. import plot
from itertools import groupby
from collections import defaultdict, namedtuple
from ..utils import cached_property, remove_suffix, get_process_pool, bucket, unzip
from ..interval import IntervalMixin, interval_pileup, get_containers
from .. import samutil
from ..samutil import parse_region
import pysam
from pysam import TabixFile, asGTF


class GenomeBrowser(object):
    def __init__(self):
        self._tracks = []

    def add_gene_track(self, gtf):
        GeneTrack(gene_models)

    def show(self, outfile):
        logging.info('Plot to %s', outfile)
        for track in self._tracks:
            track.show(args.region)


class GeneTrack(object):
    pass


class GTFRecord(namedtuple('GTFRecord', 'seqname,source,feature,start1,end,score,strand,frame,attribute')):
    """

    seqname - name of the chromosome or scaffold; chromosome names can be given with or without the 'chr' prefix. Important note: the seqname must be one used within Ensembl, i.e. a standard chromosome name or an Ensembl identifier such as a scaffold ID, without any additional content such as species or assembly. See the example GFF output below.
    source - name of the program that generated this feature, or the data source (database or project name)
    feature - feature type name, e.g. Gene, Variation, Similarity
    start - Start position of the feature, with sequence numbering starting at 1.
    end - End position of the feature, with sequence numbering starting at 1.
    score - A floating point value.
    strand - defined as + (forward) or - (reverse).
    frame - One of '0', '1' or '2'. '0' indicates that the first base of the feature is the first base of a codon, '1' that the second base is the first base of a codon, and so on..
    attribute - A semicolon-separated list of tag-value pairs, providing additional information about each feature.

    chr22   protein_coding  transcript      42095503        42195458        .       +       .       gene_id "ENSG00000167077"; transcript_id "ENST00000401548"; gene_name "MEI1"; gene
    """
    _pat_attr = re.compile('(\w+) "([^;"]*)";')

    def __new__(cls, seqname, source, feature, start1, end, score, strand, frame, attribute):
        start1 = int(start1)
        end = int(end)
        self = super(GTFRecord, cls).__new__(seqname, source, feature, start1, end, score, strand, frame, attribute)
        for match in self._pat_attr.findall(attribute):
            g = match.groups()
            g[0], g[1]

        return self

    def as_interval(self):
        return Interval(self.seqname, self.start, self.end, data=self)

    @property
    def start(self):
        return self.start1 - 1


class GTFFile(object):
    """
    +----------+----------+-------------------------------+
    |*Column*  |*Name*    |*Content*                      |
    +----------+----------+-------------------------------+
    |1         |contig    |the chromosome name            |
    +----------+----------+-------------------------------+
    |2         |feature   |The feature type               |
    +----------+----------+-------------------------------+
    |3         |source    |The feature source             |
    +----------+----------+-------------------------------+
    |4         |start     |genomic start coordinate       |
    |          |          |(0-based)                      |
    +----------+----------+-------------------------------+
    |5         |end       |genomic end coordinate         |
    |          |          |(0-based)                      |
    +----------+----------+-------------------------------+
    |6         |score     |feature score                  |
    +----------+----------+-------------------------------+
    |7         |strand    |strand                         |
    +----------+----------+-------------------------------+
    |8         |frame     |frame                          |
    +----------+----------+-------------------------------+
    |9         |attributes|the attribute field            |
    +----------+----------+-------------------------------+

    GTF formatted entries also define the following fields that
    are derived from the attributes field:

    +--------------------+------------------------------+
    |*Name*              |*Content*                     |
    +--------------------+------------------------------+
    |gene_id             |the gene identifier           |
    +--------------------+------------------------------+
    |transcript_id       |the transcript identifier     |
    +--------------------+------------------------------+
    """
    def __init__(self, gtf):  # if not seekable, load all info
        self._gtf = gtf
        self._parser = asGTF()
        self.has_tabix = os.path.exists('{0}.tbi'.format(self._gtf))

    @cached_property
    def _data(self):  # load all
        return list(iter(self))

    def __iter__(self):
        with open(self._gtf) as fp:
            for line in fp:
                yield self._parser(line)

    def _fetch(self, region):
        if not self.has_tabix:
            raise Exception('Currently, tabix is required for region query')
        with TabixFile(self._gtf, parser=self._parser) as tabix:
            for row in tabix.fetch(region):
                yield row

    def get_gene_models(self, region=None):
        if region:
            return GeneModels(self._fetch(region), region=region)
        return GeneModels(self._data)


class Transcript(IntervalMixin):
    def __init__(self, gtf_rows):
        self.contig = None
        self.start = None
        self.end = None
        self.exons = []
        self.CDSs = []
        self.UTRs = []
        self.start_codon = None
        self.stop_codon = None
        self.row = None
        self.source = None
        self.id = None
        self.gene_id = None
        self.gene_name = None
        for row in gtf_rows:
            attrs = row.asDict()
            self.id = attrs['transcript_id']
            self.gene_id = attrs['gene_id']
            if row.feature == 'transcript':
                self.row = row
                self.gene_name = attrs['gene_name']
            elif row.feature == 'exon':
                self.exons.append(row)
            elif row.feature == 'CDS':
                self.CDSs.append(row)
            elif row.feature == 'UTR':
                self.UTRs.append(row)
            elif row.feature == 'start_codon':
                self.start_codon = row
            elif row.feature == 'stop_codon':
                self.stop_codon = row

        if self.row:
            self.contig = self.row.contig
            self.start = self.row.start
            self.end = self.row.end
            self.strand = self.row.strand
            self.source = self.row.source

        # sort
        if self.exons:
            self.exons = sorted(self.exons, key=lambda x:int(x.asDict()['exon_number']))
        if self.CDSs:
            self.CDSs = sorted(self.CDSs, key=lambda x:int(x.asDict()['exon_number']))

    def __str__(self):
        return 'Transcript[{self.gene_id}({self.gene_name}):{self.id},{self.source},{self.contig}:{self.start}-{self.end}({length}),nexons:{nexon},CDS:{nCDS}]'.format(
                self=self, nexon=len(self.exons), nCDS=len(self.CDSs), length=self.end - self.start)


class GeneModels(object):
    def __init__(self, gtf_rows, region=None):
        self.region = region
        self.transcripts = trs = []

        gtf_rows = sorted(gtf_rows, key=lambda x: x.asDict()['gene_id'])
        for gene_id, rows in groupby(gtf_rows, lambda x: x.asDict()['gene_id']):
            rows = list(rows)
            tr_rows = defaultdict(list)
            gene_row = None
            for row in rows:
                if row.feature == 'gene':
                    gene_row = row
                    continue
                tr_rows[row.asDict()['transcript_id']].append(row)
            for rows in tr_rows.values():
                tr = Transcript(rows)
                trs.append(tr)

        # Reorder by start asc, length desc
        trs.sort(key=lambda x: (x.start, - x.end))
        for tr in trs:
            logging.info(tr)

    def plot(self, ax=None, text_size=2.5, gene_only=False):
        import matplotlib
        import matplotlib.pyplot as plt
        import matplotlib.patches as patches
        layouts = []
        trs = self.transcripts  # TODO filter some?
        levels = interval_pileup(trs)

        y_margin = 70
        width_CDS = 40
        width_exon = 25
        strand_styles = {'+': '>', '-': '<', '.': '--'}

        containers = get_containers(trs)
        if len(containers) != 1:
            logging.error('Cannot plot genes of multiple or zero contigs: %s', containers)
        container = tuple(containers.values())[0]
        if self.region is None:
        #matplotlib.rcParams['font.size'] = 2.  # TODO local
            xlim = (container.start, container.end)
        else:
            parsed = parse_region(self.region)
            xlim = (parsed[1], parsed[2])

        def plot_transcript(tr, level):
            y = - level * y_margin
            # basic line (TODO arrows for strand)
            ls = strand_styles[tr.strand]  # TODO
            ax.plot((tr.start, tr.end), (y, y), linestyle='-', color='b', linewidth=0.5)
            # exon
            for row in (tr.exons or []):
                ax.add_patch(patches.Rectangle((row.start, y - width_exon / 2.), row.end - row.start, width_exon, facecolor='b', linewidth=0))
            for row in (tr.CDSs or []):
                ax.add_patch(patches.Rectangle((row.start, y - width_CDS / 2.), row.end - row.start, width_CDS, facecolor='b', linewidth=0))
            #
            tmpl = '{tr.gene_name}' if gene_only else '{tr.id}({tr.gene_name})'
            ax.text((tr.start + tr.end)/2., - (level-0.5) * y_margin, tmpl.format(tr=tr), ha='center', size=text_size)

        if ax is None:
            ax = plt.gca()
        ax.axes.get_yaxis().set_visible(False)
        ax.set_xlim(*xlim)
        ax.set_ylim(- (max(levels) + 1) * y_margin, 1 * y_margin)
        for tr, level in zip(trs, levels):
            plot_transcript(tr, level)
        #--------------------
        #   -------------
        #   ------------
        #   ----------
        #browser.add_gene_track(args.gtf)   # check if seekable
        #browser.set_region(args.region)
        #browser.show(outfile)


@command.add_sub
@argument('gtf', help='need tabix')
@argument('region', help='e.g. chr1:42100000-42150000')
@argument('-o', '--output')
def gtf_plot(args):
    outfile = 'gtf_plot.{region}.pdf'.format(region=args.region)
    plot.init()
    import matplotlib.pyplot as plt
    logging.info('plot to %s', outfile)
    plt.figure(figsize=(10., 2.))
    gtf = GTFFile(args.gtf)
    gm = gtf.get_gene_models(region=args.region)
    gm.plot()
    plt.savefig(outfile)


class SAMAnalyzer(object):
    def __init__(self, fname, region):
        self.filename = fname
        self.region = region
        self.chrom, self.start, self.end = parse_region(region)
        sam = pysam.Samfile(fname)

        self.aligned = []
        self.mapq1s = []
        self.lclips = []
        self.rclips = []
        self.lmiss = []
        self.rmiss = []
        self.linvs = []
        self.rinvs = []
        for rec in sam.fetch(region=region):
            read = samutil.Read(rec)
            if read.unmapped:
                continue
            if read.lclip > 0 and read.rclip > 0:  # discard both clipped reads
                continue
            self.aligned.append(read)
            if read.mapq <= 1:
                self.mapq1s.append(read)
            if read.lclip > 0:
                self.lclips.append(read)
            if read.rclip > 0:
                self.rclips.append(read)
            if read.mate_miss:
                if read.is_left:
                    self.rmiss.append(read)
                else:
                    self.lmiss.append(read)
            if read.mate_invert:
                if read.is_left:
                    self.rinvs.append(read)
                else:
                    self.linvs.append(read)

        #self.lmiss_x = [(a.start + a.end)/2. for a in lmiss]
        #self.rmiss_x = [(a.start + a.end)/2. for a in rmiss]
        #self.linvs_x = [(a.start + a.end)/2. for a in linvs]
        #self.rinvs_x = [(a.start + a.end)/2. for a in rinvs]

    def plot_amount(self, ax=None, bin_size=1000):
        import matplotlib.pyplot as plt
        ax = ax or plt.gca()
        bins = int(math.ceil((self.end - self.start) / bin_size))
        aligned_x = [(a.start + a.end)/2. for a in self.aligned]
        mapq1s_x = [(a.start + a.end)/2. for a in self.mapq1s]
        ax.hist(aligned_x, bins=bins, range=(self.start, self.end), linewidth=0)
        ax.hist(mapq1s_x, bins=bins, range=(self.start, self.end), linewidth=0, color='silver')

    def plot_clips(self, ax=None, bin_size=1000):
        import matplotlib.pyplot as plt
        ax = ax or plt.gca()
        bins = int(math.ceil((self.end - self.start) / bin_size))
        lclips_x = [a.start for a in self.lclips]
        rclips_x = [a.end for a in self.rclips]
        ax.hist(lclips_x, bins=bins, range=(self.start, self.end), linewidth=0, color='red')
        ax.hist(rclips_x, bins=bins, range=(self.start, self.end), linewidth=0, color='green')


def get_sample_name(fname):
    name = os.path.basename(fname)
    name = remove_suffix(name, '.sam')
    name = remove_suffix(name, '.bam')
    name = remove_suffix(name, '.sorted')
    name = remove_suffix(name, '.sort')
    return name

@command.add_sub
@argument('sam', nargs='+')
@argument('-r', '--region', required=True)
@argument('--sample', nargs='+')
@argument('--gtf', help='need tabix')
@argument('--show-clips', action='store_true')
@argument('-o', '--output', default='bam_bp_plot.pdf')
@argument('--sam-per-page', default=10, type=int)
@argument('--bin-size', default=1000, type=int)
@argument('--njobs', default=8, type=int)
#@argument('--skip-flag', type=lambda x: int(x, 16), default=0x904)
def bam_bp_plot(args):
    """ Show Breakpoints

    [break_point -> read]   # position, right, left, read properties (has_another_break, aligned portion, ...)
    """
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    from matplotlib import ticker
    import seaborn as sns
    #binned_read_stream = read_stream.subscribe(binning)
    #binned_read_stream.subscribe

    samas = []
    names = []

    def task(a):
        logging.info('Analyzing %s', a)
        i, fname = a
        sample = args.sample[i] if args.sample else get_sample_name(fname)
        sama = SAMAnalyzer(fname, region=args.region)
        return sama, sample

    #pool = get_process_pool(args.njobs)
    #for sama, sample in pool.imap(task, enumerate(args.sam), 1):
    for sama, sample in map(task, enumerate(args.sam)):
        samas.append(sama)
        names.append(sample)

    #for i, fname in enumerate(args.sam):

    chrom, start, end = parse_region(args.region)
    logging.info('Plotting %s:%s-%s', chrom, start+1, end)

    zoom = 1
    matplotlib.rcParams['figure.figsize'] = (11.7 * zoom, 8.3 * zoom)
    matplotlib.rcParams['font.size'] = 8.
    matplotlib.rcParams['axes.titlesize'] = 'medium'
    matplotlib.rcParams['axes.labelsize'] = 'medium'
    matplotlib.rcParams['axes.labelpad'] = 8.
    matplotlib.rcParams['xtick.labelsize'] = 'small'
    matplotlib.rcParams['ytick.labelsize'] = 'small'
    matplotlib.rcParams['legend.fontsize'] = 'small'

    #counters = ReadCountGenerator(sam, chrom, start=start, end=end, window=args.window, skip_flag=args.skip_flag, mass='overlap')
    #tab = pd.DataFrame([getter(counter) for counter in counters], columns=attrs)
    #tab['mid'] = (tab['start'] + tab['end']) / 2.
    #tab['density'] = 1. * tab['covlen'] / window   # like depth
    #tab['density_mapq1'] = 1. * tab['covlen_mapq1'] / window   # like depth

    #plt.tight_layout()
    #plt.title('{0}:{1}-{2}', chrom, start, end)
    # tracks
    def plot_page(samas, names):
        track_n = track_per_sample * len(samas)
        ratios = [1] * (track_per_sample * len(samas))

        if args.gtf:
            track_n += 1
            ratios.append(3)

        # TODO gridspec
        fig, sub_axs = plt.subplots(track_n, sharex=True, gridspec_kw = {'height_ratios': ratios})
        fig.subplots_adjust(hspace=0)

        sub_axs = iter(sub_axs)
        bin_size = args.bin_size
        yrot = 0
        for sama, name in zip(samas, names):
            logging.info('Plotting %s', sama.filename)
            ax = next(sub_axs)
            sama.plot_amount(ax=ax, bin_size=bin_size)
            ax.set_ylabel('#Reads\n{0}'.format(name), rotation=yrot, ha='right')
            if args.show_clips:
                ax = next(sub_axs)
                sama.plot_clips(ax=ax, bin_size=bin_size)
                ax.set_ylabel('#Clips\n{0}'.format(name), rotation=yrot, ha='right')

        #for i, (ax, x) in enumerate(zip(sub_axs, bp_xs), 1):
        #    #plt.subplot(n, 1, i, sharex=True)
        #    ax.hist(x, bins=bins, range=(start, end), linewidth=0)
            #ax = plt.gca()
            #ax.xaxis.set_major_formatter(ticker.NullFormatter())
            #ax.xaxis.set_major_locator(ticker.NullLocator())
            #ax.xaxis.set_major_locator(ticker.LinearLocator())

        if args.gtf:
            ax = next(sub_axs)
            gtf = GTFFile(args.gtf)
            gm = gtf.get_gene_models(region=args.region)
            gm.plot(ax=ax, text_size=6.5, gene_only=True)
            ax = plt.gca()
            ax.set_ylabel('Genes', rotation=yrot)

        ax = plt.gca()
        #ax.xaxis.set_major_formatter(ticker.ScalarFormatter())
        #ax.xaxis.set_major_locator(ticker.LinearLocator())
        if start is not None and end is not None:
            ax.set_xlabel('{0}:{1}-{2} ({3})'.format(chrom, start+1, end, (end - start)))
        else:
            ax.set_xlabel('{0}'.format(chrom))

    track_per_sample = 1
    if args.show_clips:
        track_per_sample += 1
    from matplotlib.backends.backend_pdf import PdfPages
    pp = PdfPages(args.output)

    for grp in bucket(zip(samas, names), args.sam_per_page):
        samas, names = unzip(grp, 2)
        plot_page(samas, names)
        pp.savefig(plt.gcf())
    pp.close()



import os
import sys
import glob
from setuptools import setup, find_packages
from setuptools.extension import Extension
#from distutils.core import setup

REQUIRES = ['argtools>=0.1.4', 'PyVCF>=0.6.8', 'cyvcf2', 'numpy', 'matplotlib', 'pandas>=0.19', 'sklearn', 'statsmodels', 'multiprocess', 'joblib', 'future', 'xlsxwriter']
REQUIRES.extend(['cython>=0.22',  'pysam>=0.9.1.4'])
if sys.version_info >= (2, 7):
    REQUIRES.append('seaborn')

try:
    from collections import OrderedDdict
except ImportError:
    REQUIRES.append('ordereddict')
try:
    from collections import Counter
except ImportError:
    REQUIRES.append('counter')

README = os.path.join(os.path.dirname(__file__), 'README.md')
if os.path.exists(README):
    long_description = open(README).read() + '\n\n'
else:
    long_description = 'A collection of algorithms and tools for genome analysis'
VERSION = '0.0.6.149'

import numpy as np
if 'build_ext' in sys.argv:
    from Cython.Distutils import build_ext
    from Cython.Compiler import Options
    from Cython.Build import cythonize
    Options.fast_fail = True
    exts = (
            cythonize('galgo/covfit/*.pyx')
         + cythonize('galgo/alignment/*.pyx')
         + [Extension('galgo.sselogsumexp', sources=['galgo/sselogsumexp.pyx', 'src/logsumexp.c'], depends=glob.glob('src/*.h'), extra_compile_args=['-msse2'])]
    )
    cmd_class = {'build_ext': build_ext}
else:
    exts = [
            Extension('galgo.covfit.pmm', sources=['galgo/covfit/pmm.c']),
            Extension('galgo.covfit.gmm_v2', sources=['galgo/covfit/gmm_v2.c']),
            Extension('galgo.alignment.basics', sources=['galgo/alignment/basics.c']),
            Extension('galgo.sselogsumexp', sources=['galgo/sselogsumexp.c', 'src/logsumexp.c'], depends=glob.glob('src/*.h'), extra_compile_args=['-msse2'])
            ]
    cmd_class = {}

exclude = ['galgo.__main__']
setup(name='galgo',
      version=VERSION,
      install_requires=REQUIRES,
      description='A collection of algorithms and tools for genome analysis',
      long_description=long_description,
      classifiers=['Programming Language :: Python',
                   'Topic :: Scientific/Engineering :: Bio-Informatics'],
      keywords='genome bio-informatics fasta',
      author='Takahiro Mimori',
      author_email='takahiro.mimori@gmail.com',
      #py_modules=['galgo'],
      package_data = {'': ['README.md']},
      #package_dir = {'galgo': 'galgo'},
      packages=find_packages(exclude=exclude),
      #scripts = ['bin/galgo'],
      scripts = ['scripts/bed2chain.py'],
      entry_points={'console_scripts': ['galgo = galgo.__main__:main']},
      url='https://github.com/m1m0r1/galgo',
      cmdclass = cmd_class,
      include_dirs = [np.get_include(), 'src'],
      ext_modules = exts,
)



#!/usr/bin/env python
from __future__ import print_function
from argtools import command, argument
import logging
import numpy as np
import itertools
from collections import namedtuple
from itertools import groupby
from multiprocess import Pool
from scipy.stats import norm
from ..covfit import load_gender_info, get_ploidies
from ..covfit import pmm
from ..utils import iter_tabs, Counter, safediv


@command.add_sub
@argument('coverage_bed')
@argument('--col-region-id')
@argument('--col-sample-id', default='sampleid')
@argument('--col-count', default='valid_count')
@argument('--col-exp-count', default='valid_exp_count')
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('-m', '--max-copy', default=10)
@argument('-la', type=float, default=0.3, help='penalty of offset')
@argument('-lb', type=float, default=1.0, help='penalty of ratio')
@argument('-a', type=float, default=0., help='penalty of offset')
@argument('-b', type=float, default=1.0, help='penalty of ratio')
@argument('-f', '--max-diff', type=float, default=1e-6, help='max diff of log likelihood')
@argument('-p', '--process', type=int, default=8)
@argument('--step-limit', type=int, help='the number of steps until inference stopped')
@argument('--gender-info', help='col1: sample_id, col2: {1:male, 2:female}')
def covfit_pmm(args):
    """

    1. chrom
    2. start
    3. end
    4. region id
    """
    sample_genders = load_gender_info(args.gender_info)
    zoom = 2. / args.diploid_value

    def run():
        param_keys = 'status top_cn top_theta a b c theta ia ib ic ll nsamples step'.split(' ')
        pool = Pool(args.process)
        #logging.info(sample_genders)

        task_inputs = gen_inputs()

        if args.process == 1:
            results = itertools.imap(task, task_inputs)
        else:
            results = pool.imap(task, task_inputs, 100)

        data_header = ['chrom', 'start', 'end', 'region_id', 'nsamples']
        param_header = ['status', 'step', 'ia', 'ib', 'a', 'b', 'top_cn', 'top_theta', 'theta', 'll']
        out_header = data_header + param_header
        print (*out_header, sep='\t')
        for data, params in results:
            row = [data[c] for c in data_header] \
                + [params[c] for c in param_header]
            print (*row, sep='\t')


    def gen_inputs():
        with open(args.coverage_bed) as fp:
            it = iter_tabs(fp)
            header = next(it)
            idx_region_id = header.index(args.col_region_id) if args.col_region_id else 3
            idx_sample_id = header.index(args.col_sample_id)
            idx_count = header.index(args.col_count)
            idx_exp_count = header.index(args.col_exp_count)

            for region_id, rows in groupby(it, lambda x: x[idx_region_id]):
                rows = list(rows)
                chrom = rows[0][0]

                samples = []
                genders = []
                counts = []
                exp_counts = []

                for row in rows:
                    sample_id = row[idx_sample_id]
                    count = float(row[idx_count])
                    exp_count = float(row[idx_exp_count]) / zoom
                    raw_cn = safediv(count, exp_count)
                    if np.isnan(raw_cn):
                        continue
                    if raw_cn > args.max_copy * 1.1:   # regard this as outlier
                        logging.warning('Omit outlier (data: %s)', zip(header, row))
                        continue
                    samples.append(sample_id)
                    genders.append(sample_genders[sample_id])
                    counts.append(count)
                    exp_counts.append(exp_count)

                ploidies = get_ploidies(chrom, genders)

                data = dict(
                        chrom = rows[0][0],
                        start = int(rows[0][1]),
                        end = int(rows[0][2]),
                        region_id = region_id,
                        counts = counts,
                        exp_counts = exp_counts,
                        la = args.la,
                        lb = args.lb,
                        a = args.a,
                        b = args.b,
                        max_diff = args.max_diff,
                        step_limit = args.step_limit,
                        ploidies = ploidies,
                        max_copy = args.max_copy,
                        nsamples = len(ploidies),
                )

                yield data

    def task(data):
        chrom = data['chrom']
        start = data['start']
        end = data['end']
        name = data['region_id']
        ploidies = data['ploidies']
        init_a = data['a']
        init_b = data['b']
        la = data['la']
        lb = data['lb']
        counts = np.array(data['counts'])
        exp_counts = np.array(data['exp_counts'])
        max_diff = data['max_diff']
        step_limit = data['step_limit']
        max_copy = data['max_copy']

        n_callable = len(counts)
        logging.info('[%s] n_callable: %s', name, n_callable)
        results = []
        norm = None
        try:
            assert n_callable > 0, 'The number of callable samples ({0}) is less than the minimum value'.format(n_callable)
            norm = pmm.Normalizer(counts=counts, exp_counts=exp_counts, ploidies=ploidies, max_copy=max_copy, a=init_a, b=init_b, la=la, lb=lb, max_diff=max_diff, step_limit=step_limit, name=name)
            result = norm.iteration()
            top_cn = np.argmax(result.theta)
            params = {
                    'status': 1,
                    'ia': init_a,
                    'ib': init_b,
                    'a': result.a,
                    'b': result.b,
                    'theta': ','.join(map(str, result.theta)),
                    'step': result.step,
                    'top_cn': top_cn,
                    'top_theta': result.theta[top_cn],
                    'll': result.ll,
            }
        except AssertionError as e:
            logging.warning('Assert error: %s', e)
            logging.warning('Cannot optimize model for %s', data)
            step = 0 if norm is None else norm.step
            params = {
                    'status': 0,
                    'ia': init_a,
                    'ib': init_b,
                    'a': init_a,
                    'b': init_b,
                    'theta': '',
                    'step': step,
                    'top_cn': 0,
                    'top_theta': 0,
                    'll': 0,
            }
        results.append(params)

        logging.info('Emitting %s:%s-%s, %s, %s', chrom, start, end, name, params)
        return (data, params)

    run()


@command.add_sub
@argument('fit_result')
@argument('--min-call-nsample', type=int, default=1)
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('-t', '--threshold', type=float, default=.8)
@argument('--sex-info', help='list file of 1:male, 2:female')
def covfit_pmm_genotype(args):
    """
    """
    nsample = args.nsample
    zoom = 2. / args.diploid_value
    gender_info = [int(line.rstrip()) for line in open(gender_info)]

    for row in iter_tabs(open(args.fit_result)):
        chrom = row[0]
        start = int(row[1])
        end = int(row[2])
        name = row[3]

        depths = np.array(row[-nsample:], dtype=float)
        callable_idxs = np.where(~ np.isnan(depths))[0]
        n_callable = len(callable_idxs)

        mod_depths = (depths * zoom)[callable_idxs]

        #mod_depths = depths * zoom
        keys = row[-args.nsample-2].split(':')
        params = dict(zip(keys, row[-args.nsample-1].split(':')))
        a = float(params['a'])
        b = float(params['b'])
        c = float(params['c'])
        try:
            assert n_callable >= args.min_call_nsample, 'The number of callable samples ({0}) is less than the minimum value'.format(n_callable)
            theta = np.array(map(float, params['theta'].split(',')))
            max_copy = len(theta) - 1

            ploidies = tuple(np.array(get_ploidies(chrom))[callable_idxs])

            #ploidies = get_ploidies(chrom)
            #ploidies = np.array(ploidies)

            norm = em.Normalizer(mod_depths, ploidies=ploidies, max_copy=max_copy, a=a, b=b, c=c, theta=theta, name=name)
            gts = list(norm.iter_genotypes())
            gts1 = []   # genotyped copy numbers
            gts2 = []   # genotype probability of 0/1/2
            for prob in gts:
                max_cn = np.argmax(prob)
                if prob[max_cn] >= args.threshold:
                    gts1.append(max_cn)
                else:
                    gts1.append('.')
                gts2.append('{0:.1f}:{1:.1f}:{2:.1f}'.format(prob[0], prob[1], prob[2]))
        except (ValueError, AssertionError):
            gts1 = ['.'] * nsample
            gts2 = ['.:.:.'] * nsample

        out_row = list(row[:-args.nsample-2])
        out_row.append(a)
        out_row.append(b)
        out_row.append(c)
        out_row.append(params['theta'])
        out_row.append(','.join(map(str, gts1)))
        out_row.append(','.join(map(str, gts2)))
        print (*out_row, sep='\t')



#!/usr/bin/env python
from __future__ import print_function
from argtools import command, argument
import itertools
from operator import itemgetter, attrgetter
from collections import namedtuple

ChainHeader = namedtuple('ChainHeader', 'score tName tSize tStrand tStart tEnd qName qSize qStrand qStart qEnd id')


def bed_to_chains(bed, rlengths, inverse=False):
    """ Convert bed to chain file

    Args:
        bed: iterable of (chrom, start, end, length)
        rlenths : dictionary of {rname: length}
    Yields:
        {'header': header tokens
         'data': [(size, dt, dq), ... , (size,)}

    See http://genome.ucsc.edu/goldenPath/help/chain.html for more details.

    header tokens:
        score -- chain score
        tName -- chromosome (reference sequence)
        tSize -- chromosome size (reference sequence)
        tStrand -- strand (reference sequence)
        tStart -- alignment start position (reference sequence)
        tEnd -- alignment end position (reference sequence)
        qName -- chromosome (query sequence)
        qSize -- chromosome size (query sequence)
        qStrand -- strand (query sequence)
        qStart -- alignment start position (query sequence)
        qEnd -- alignment end position (query sequence)
        id -- chain ID

    data tokens (the end of data token contains only first token):
        size -- the size of the ungapped alignment
        dt -- the difference between the end of this block and the beginning of the next block (reference sequence)
        dq -- the difference between the end of this block and the beginning of the next block (query sequence)

    >>> recs = [
    ...    LengthBed('chr1', 1000, 1001, 1),
    ...    LengthBed('chr1', 1002, 1002, 4),
    ...    LengthBed('chr1', 1010, 1010, 2),
    ...
    ...    LengthBed('chr2', 1000, 1010, 0),  # chrom changed
    ...    LengthBed('chr2', 1020, 1021, 1),
    ...    LengthBed('chr2', 1030, 1030, 8),
    ...    LengthBed('chr2', 1040, 1043, 8),
    ...    LengthBed('chr2', 2000, 3000, 3000),
    ...    LengthBed('chr2', 3500, 4000, 0),
    ... ]
    >>> rlengths = {'chr1': 10000, 'chr2': 30000, 'chr3': 10000}

    >>> chains = list(bed_to_chains(recs, rlengths))
    >>> attrgetter('score', 'tName', 'tSize', 'tStart', 'tEnd', 'qName', 'qSize', 'qStart', 'qEnd', 'id')(chains[0]['header'])
    (0, 'chr1', 10000, 0, 10000, 'chr1', 10006, 0, 10006, 1)
    >>> chains[0]['data']
    [(1002, 0, 4), (8, 0, 2), (8990,)]

    >>> attrgetter('score', 'tName', 'tSize', 'tStart', 'tEnd', 'qName', 'qSize', 'qStart', 'qEnd', 'id')(chains[1]['header'])
    (0, 'chr2', 30000, 0, 30000, 'chr2', 31503, 0, 31503, 2)
    >>> chains[1]['data']
    [(1000, 10, 0), (20, 0, 8), (10, 3, 8), (957, 1000, 3000), (500, 500, 0), (26000,)]

    >>> attrgetter('score', 'tName', 'tSize', 'tStart', 'tEnd', 'qName', 'qSize', 'qStart', 'qEnd', 'id')(chains[2]['header'])
    (0, 'chr3', 10000, 0, 10000, 'chr3', 10000, 0, 10000, 3)
    >>> chains[2]['data']
    [(10000,)]

    # inverse mode
    >>> chains = list(bed_to_chains(recs, rlengths, inverse=True))
    >>> attrgetter('score', 'tName', 'tSize', 'tStart', 'tEnd', 'qName', 'qSize', 'qStart', 'qEnd', 'id')(chains[0]['header'])
    (0, 'chr1', 10006, 0, 10006, 'chr1', 10000, 0, 10000, 1)
    >>> chains[0]['data']
    [(1002, 4, 0), (8, 2, 0), (8990,)]
    """
    rname_seen = set()
    chain_id = 1
    for (rname, recs) in itertools.groupby(bed, itemgetter(0)):
        assert rname not in rname_seen, "svbed are not sorted in rname! ({0} was appeared again)".format(rname)
        rname_seen.add(rname)
        tSize = qSize = rlengths[rname]
        data = []

        last_pos = 0
        for rec in recs:
            common_len = rec.start - last_pos
            ref_len = rec.end - rec.start
            alt_len = rec.length
            if ref_len == alt_len: # uncount the feature with same length
                continue
            elif ref_len > 0 or alt_len > 0:
                data.append((common_len, ref_len, alt_len))
                last_pos = rec.end
                qSize += (alt_len - ref_len)
            else:
                continue

        last_len = tSize - last_pos
        if inverse:
            tSize, qSize = qSize, tSize
            data = [(common_len, alt_len, ref_len) for common_len, ref_len, alt_len in data]

        data.append((last_len,))

        yield {'header': ChainHeader(0, rname, tSize, '+', 0, tSize,
                                        rname, qSize, '+', 0, qSize, chain_id),
               'data': data}

        chain_id += 1

    # remaining
    for rname in set(rlengths) - rname_seen:
        assert rname not in rname_seen
        rname_seen.add(rname)
        tSize = qSize = rlengths[rname]
        data = []
        last_pos = 0
        last_len = tSize - last_pos
        data.append((last_len,))
        yield {'header': ChainHeader(0, rname, tSize, '+', 0, tSize,
                                        rname, qSize, '+', 0, qSize, chain_id),
               'data': data}

        chain_id += 1


def iter_tabs(fname):
    with open(fname) as fp:
        for line in fp:
            row = line.rstrip('\r\n').split('\t')
            yield row

def read_genomes(fname):
    rlens = []
    for row in iter_tabs(fname):
        rlens.append((row[0], int(row[1])))
    return rlens

LengthBed = namedtuple('Bed', 'chrom start end length')
def iter_beds(fname, length_column=4):
    idx = length_column - 1
    for row in iter_tabs(fname):
        yield LengthBed(row[0], int(row[1]), int(row[2]), int(row[idx]))

@command
@argument('bed', help='bed file with "region replaced length"')
@argument('genomes', help='tsv file of rname and length as 1st and 2nd column, respectively (e.g. .fai file)')
@argument('-c', '--column', default=4, help='the column of "region replaced length" in bed')
@argument('-i', '--inverse', action='store_true')
def bed2chain(args):
    """ Convert bed file into chain file
    """
    rlens = read_genomes(args.genomes)

    # print chain file
    it = iter_beds(args.bed)
    for chains in bed_to_chains(it, dict(rlens), inverse=args.inverse):
        print('chain', *chains['header'], sep=' ')
        for row in chains['data']:
            print(*row, sep='\t')
        print()

if __name__ == '__main__':
    command.run()



from __future__ import absolute_import
from __future__ import print_function
import sys
import logging
from argtools import command, argument
#import vcf as VCF
import cyvcf2 as VCF
import pandas as pd
from io import open
from ..plot import PdfGrid
from ..utils import safediv
import numpy as np
from scipy import stats
from sklearn.linear_model import LinearRegression
from sklearn.feature_selection import f_regression


def read_sample_file(fname):
    """
    ID_1 ID_2 missing cov_1 cov_2 cov_3 cov_4 pheno1 bin1
    0 0 0 D D C C P B
    1 1 0.007 1 2 0.0019 -0.008 1.233 1
    2 2 0.009 1 2 0.0022 -0.001 6.234 0
    3 3 0.005 1 2 0.0025 0.0028 6.121 1
    4 4 0.007 2 1 0.0017 -0.011 3.234 1
    5 5 0.004 3 2 -0.012 0.0236 2.786 0
    """
    default_types = {'sex': 'D', 'age': 'C'}
    with open(fname) as fp:
        header = next(fp).rstrip().split(' ')
        types = next(fp).rstrip().split(' ')

        types = [default_types.get(col, t) for col, t in zip(header, types)]  # override with default types
        dtypes = {'ID_1': 'str', 'ID_2': 'str', 'missing': 'float'}
        for col, t in zip(header, types):
            if t == '0':
                continue
            elif t in ('C', 'P'):
                dtypes[col] = 'float'
            elif t in ('D', 'B'):
                dtypes[col] = 'category'

        tab = pd.read_table(fp, sep=' ', names=header, dtype=dtypes)
    return tab


def get_vcf_target_rec(vcf, id):
    reader = VCF.Reader(vcf)
    for rec in reader:
        if rec.ID == id:
            return reader, rec

def isnan(obj):
    return isinstance(obj, float) and np.isnan(obj)


@command.add_sub
@argument('sample')
@argument('-f', '--vcf', help='VCF or gzipped VCF')
@argument('-i', '--id')
@argument('-o', '--output', default='test.pdf')
def pheno_plot(args):
    tab = read_sample_file(args.sample)

    min_qual = 90
    col_pheno = tab.columns[-1]  # default target pheno

    import matplotlib as mp
    mp.use('Agg')
    import matplotlib.pyplot as plt
    import seaborn as sns

    def with_gt_dosage(tab, vcf, id):
        vcf_reader, rec = get_vcf_target_rec(vcf, id)
        if not rec:
            logging.error('Cannot find {} in vcf: {}'.format(id, vcf))
            raise Exception()

        samples = vcf_reader.samples
        if rec.INFO.get('SVTYPE') == 'CNV':
            dosage = rec.format('CN')[:, 0]
            quality = rec.format('CNQ')[:, 0]  # 0-100
            logging.info('Variant is CNV')
        else:
            raise NotImplementedError

        dosage_tab = pd.DataFrame({'id': samples, 'dosage': dosage, 'dosage_qual': quality})
        tab = tab.merge(dosage_tab, left_on='ID_1', right_on='id', how='left')
        return tab

    if args.vcf and args.id:
        tab = with_gt_dosage(tab, vcf=args.vcf, id=args.id)
    logging.info(tab)

    # TODO category phenotype
    def pheno_plot(tab):
        # SNP: 0, 1, 2
        # CN: 0, 1, 2, ...
        tab[col_pheno].hist()

    def pheno_sex_plot(tab):
        sns.violinplot(x=tab['sex'], y=tab[col_pheno])

    def pheno_sex_age_plot(tab):
        plt.scatter(x=tab['age'], y=tab[col_pheno], c=tab['sex'])

    def pheno_sex_dosage_plot(tab):
        # SNP: 0, 1, 2
        # CN: 0, 1, 2, ...
        #plt.scatter(x=tab['dosage'], y=tab[col_pheno], c=tab['sex'])
        sns.swarmplot(x=tab['dosage'], y=tab[col_pheno], hue=tab['sex'])

    logging.info('Plot to %s', args.output)
    plt.figure(figsize=(11.7, 8.3))
    with PdfGrid(args.output, row=2, col=2) as pg:
        plt.gcf().subplots_adjust()
        next(pg)
        pheno_plot(tab)
        next(pg)
        pheno_sex_plot(tab)
        next(pg)
        pheno_sex_age_plot(tab)
        next(pg)
        pheno_sex_dosage_plot(tab)



class VCFAssocCNV(object):
    _header = ['chrom', 'start', 'end', 'name', 'nsample', 'ncalled', 'call_rate', 'alt_sample', 'alt_rate', 'nassoc']

    def __init__(self, vcf, sample, min_qual=90, sample_id='ID_1', pheno_id=None, covariates=None, sex='both'):
        self.min_qual = min_qual
        sample_tab = read_sample_file(sample)
        if pheno_id is None:
            pheno_id = sample_tab.columns[-1]   # take last column as assoc phenotype
        self.pheno_id = pheno_id
        self.covariates = covariates or []

        self._reader = reader = VCF.Reader(vcf)
        vcf_samples = reader.samples

        # TODO restrict sex
        if sex == 'both':
            pass
        elif sex == 'male':
            sample_tab = sample_tab[sample_tab.sex == '1']
        elif sex == 'female':
            sample_tab = sample_tab[sample_tab.sex == '2']
        else:
            logging.warning('sex is one of 0/1/2')
            raise NotImplementedError

        assoc_tab = pd.DataFrame({'vcf_sample': vcf_samples})
        assoc_tab = assoc_tab.merge(sample_tab, left_on='vcf_sample', right_on=sample_id, how='left')
        self._assoc_tab = assoc_tab
        self.sample_ok = assoc_tab[sample_id].map(lambda x: not isnan(x))   # ID_1 is exist in sample file
        self.pheno_ok = assoc_tab[pheno_id].map(lambda x: not isnan(x))   # phenotype is not an NA
        if covariates:
            self.covs_ok = np.prod([assoc_tab[cov_id].map(lambda x: not isnan(x)) for cov_id in self.covariates], dtype=bool, axis=0)  # covariates are not an NA
        else:
            self.covs_ok = np.ones(len(assoc_tab), dtype=bool)
        #print (self.covs_ok)

    def iter_assocs(self):
        for rec in self._reader:
            arec = self._assoc(rec)
            if arec:
                yield arec

    def _get_data(self, rec):
        if rec.INFO.get('SVTYPE') != 'CNV':
            return

        data = {}
        data['chrom'] = rec.CHROM
        data['start'] = rec.POS - 1
        data['end'] = rec.INFO['END']
        data['name'] = rec.ID
        data['nsample'] = self.sample_ok.sum()

        dosage = pd.Series(rec.format('CN')[:, 0])
        quality = pd.Series(rec.format('CNQ')[:, 0])  # 0-100
        is_called = dosage.map(lambda x: not isnan(x))
        has_good_call = is_called & (quality >= self.min_qual)

        total_call = np.sum(self.sample_ok & is_called)
        valid_call = has_good_call & self.sample_ok
        valid_call_nsample = valid_call.sum()
        data['ncalled'] = valid_call_nsample

        alt_sample = (dosage[valid_call] != 2).sum()
        data['alt_sample'] = alt_sample
        data['alt_rate'] = safediv(1. * alt_sample, valid_call_nsample, error=0.)
        data['call_rate'] = safediv(1. * valid_call_nsample, data['nsample'], error=0.)

        is_assoc_target = has_good_call & self.sample_ok & self.pheno_ok & self.covs_ok
        data['a_dosage'] = a_dosage = dosage[is_assoc_target]
        data['a_covs'] = self._assoc_tab[self.covariates][is_assoc_target] if self.covariates else None
        data['a_pheno'] = a_pheno = self._assoc_tab[self.pheno_id][is_assoc_target]
        data['nassoc'] = len(a_dosage)

        return data


class VCFAssocCNVPearson(VCFAssocCNV):
    @property
    def header(self):
        return self._header + ['cor', 'pvalue']

    def _assoc(self, rec):
        data = self._get_data(rec)
        if data is None:
            return

        if data['nassoc']:
            #print (a_dosage, a_pheno)
            cor, pvalue = stats.pearsonr(data['a_dosage'], data['a_pheno'])
        else:
            cor, pvalue = float('nan'), float('nan')

        data['cor'] = cor
        data['pvalue'] = pvalue
        return data


def hierarchical_regression(X, y):
    import statsmodels.api as sm
    X = X.astype(float)
    X = sm.add_constant(X)
    results = []
    pvalues = []
    for icol in range(X.shape[1]):
        X1 = X[:, :icol + 1]
        res = sm.OLS(y, X1).fit()   # TODO normalize?
        if results:
            lr_test, p_value, df_diff = res.compare_lr_test(results[-1])
        else:
            p_value = res.pvalues[0]  # pvalue for intercept
        pvalues.append(p_value)
        results.append(res)

    return {'results': results, 'pvalues': pvalues}


class VCFAssocCNVLinear(VCFAssocCNV):
    @property
    def header(self):
        exts = []
        exts.append('b0')
        for cov_id in self.covariates:
            exts.append('cov_b_{0}'.format(cov_id))
            exts.append('cov_p_{0}'.format(cov_id))
        exts.append('beta')
        exts.append('pvalue')
        return self._header + exts

    def _assoc(self, rec):
        data = self._get_data(rec)
        if data is None:
            return

        y = data['a_pheno']
        if len(data['a_covs']):
            X = np.c_[data['a_covs'], data['a_dosage'].reshape((-1, 1))]
            #X = np.c_[data['a_dosage'].reshape((-1, 1)), data['a_covs']]
        else:
            X = np.c_[data['a_dosage']].reshape((-1, 1))

        #logging.info(X)
        lin_reg= LinearRegression(normalize=True)
        if len(X):
            res = hierarchical_regression(X, y.values)
            b0 = res['results'][-1].params[0]
            theta = res['results'][-1].params[1:]
            pvals = res['pvalues'][1:]  # ignore intercept
            # 
            # lin_reg.fit(X, y)
            # b0 = lin_reg.intercept_
            # theta = lin_reg.coef_
            # F, pvals = f_regression(X, y, center=True)
        else:
            b0 = float('nan')
            theta = [float('nan')] * (len(self.covariates) + 1)
            pvals = [float('nan')] * (len(self.covariates) + 1)

        data['b0'] = b0
        for i, cov_id in enumerate(self.covariates): #theta[1:]):
            data['cov_b_{0}'.format(cov_id)] = theta[i]
            data['cov_p_{0}'.format(cov_id)] = pvals[i]
        data['beta'] = theta[-1]
        data['pvalue'] = pvals[-1]
        return data


@command.add_sub
@argument('sample')
@argument('vcf', help='VCF or gzipped VCF')
@argument('-m', '--method', choices=['pearson', 'spearman', 'linear', 't', 'welch', 'logistic'])
@argument('-s', '--sex', choices=['both', 'male', 'female'], default='both')
@argument('-p', '--pheno-id')
@argument('-c', '--covs')
def vcf_assoc_cnv(args):
    """
    Assume autosomals

    Output TSV:
        chrom
        start
        end
        name
        (regino_type, strand, region_id, region_name) <= need gene annotation file

        # basic sample info
        call_rate : change with min_cnq
        alt_sample :
        alt_rate :
        cns :
        nsample

        # pearson (cor)
        cor
        pvalue

        # linear (standardized linear)
        b0
        cov_b_{cov_name1}
        cov_p_{cov_name1}
        cov_b_{cov_name2}
        cov_p_{cov_name2}
        beta                # genotype
        pvalue

        # logistic regression  # TODO
    """
    vcffile = '/dev/stdin' if args.vcf == '-' else args.vcf
    min_cnq = 90

    cls = {
          'pearson': VCFAssocCNVPearson,
          'linear': VCFAssocCNVLinear,
    }[args.method]

    covs = [] if not args.covs else args.covs.split(',')
    vcf_assoc = cls(args.vcf, args.sample, sex=args.sex, pheno_id=args.pheno_id, covariates=covs)

    header = vcf_assoc.header
    print (*header, sep='\t')
    for rec in vcf_assoc.iter_assocs():
        print (*[rec[col] for col in header], sep='\t')
